use nom_sql::{ArithmeticExpression, ColumnSpecification, OrderType};
use petgraph::graph::NodeIndex;
use std::cell::RefCell;
use std::fmt::{Debug, Display, Error, Formatter};
use std::rc::Rc;

use column::Column;
use common::DataType;
use dataflow::ops;
use dataflow::ops::filter::FilterCondition;
use dataflow::ops::grouped::aggregate::Aggregation as AggregationKind;
use dataflow::ops::grouped::extremum::Extremum as ExtremumKind;
use std::collections::HashMap;
use {FlowNode, MirNodeRef};

/// Helper enum to avoid having separate `make_aggregation_node` and `make_extremum_node` functions
pub enum GroupedNodeType {
    Aggregation(ops::grouped::aggregate::Aggregation),
    Extremum(ops::grouped::extremum::Extremum),
    GroupConcat(String),
}

pub struct MirNode {
    pub name: String,
    pub from_version: usize,
    pub columns: Vec<Column>,
    pub inner: MirNodeType,
    pub ancestors: Vec<MirNodeRef>,
    pub children: Vec<MirNodeRef>,
    pub flow_node: Option<FlowNode>,
}

impl MirNode {
    pub fn new(
        name: &str,
        v: usize,
        columns: Vec<Column>,
        inner: MirNodeType,
        ancestors: Vec<MirNodeRef>,
        children: Vec<MirNodeRef>,
    ) -> MirNodeRef {
        let mn = MirNode {
            name: String::from(name),
            from_version: v,
            columns,
            inner,
            ancestors: ancestors.clone(),
            children: children.clone(),
            flow_node: None,
        };

        let rc_mn = Rc::new(RefCell::new(mn));

        // register as child on ancestors
        for ancestor in &ancestors {
            ancestor.borrow_mut().add_child(rc_mn.clone());
        }

        rc_mn
    }

    /// Adapts an existing `Base`-type MIR Node with the specified column additions and removals.
    pub fn adapt_base(
        node: MirNodeRef,
        added_cols: Vec<&ColumnSpecification>,
        removed_cols: Vec<&ColumnSpecification>,
    ) -> MirNodeRef {
        let over_node = node.borrow();
        match over_node.inner {
            MirNodeType::Base {
                ref column_specs,
                ref keys,
                ..
            } => {
                let new_column_specs: Vec<(ColumnSpecification, Option<usize>)> = column_specs
                    .iter()
                    .cloned()
                    .filter(|&(ref cs, _)| !removed_cols.contains(&cs))
                    .chain(
                        added_cols
                            .iter()
                            .map(|c| ((*c).clone(), None))
                            .collect::<Vec<(ColumnSpecification, Option<usize>)>>(),
                    )
                    .collect();
                let new_columns: Vec<Column> = new_column_specs
                    .iter()
                    .map(|&(ref cs, _)| Column::from(&cs.column))
                    .collect();

                assert_eq!(
                    new_column_specs.len(),
                    over_node.columns.len() + added_cols.len() - removed_cols.len()
                );

                let new_inner = MirNodeType::Base {
                    column_specs: new_column_specs,
                    keys: keys.clone(),
                    adapted_over: Some(BaseNodeAdaptation {
                        over: node.clone(),
                        columns_added: added_cols.into_iter().cloned().collect(),
                        columns_removed: removed_cols.into_iter().cloned().collect(),
                    }),
                };
                MirNode::new(
                    &over_node.name,
                    over_node.from_version,
                    new_columns,
                    new_inner,
                    vec![],
                    over_node.children.clone(),
                )
            }
            _ => unreachable!(),
        }
    }

    /// Wraps an existing MIR node into a `Reuse` node.
    /// Note that this does *not* wire the reuse node into ancestors or children of the original
    /// node; if required, this is the responsibility of the caller.
    pub fn reuse(node: MirNodeRef, v: usize) -> MirNodeRef {
        let rcn = node.clone();

        let mn = MirNode {
            name: node.borrow().name.clone(),
            from_version: v,
            columns: node.borrow().columns.clone(),
            inner: MirNodeType::Reuse { node: rcn },
            ancestors: vec![],
            children: vec![],
            flow_node: None, // will be set in `into_flow_parts`
        };

        Rc::new(RefCell::new(mn))
    }

    pub fn can_reuse_as(&self, for_node: &MirNode) -> bool {
        let mut have_all_columns = true;
        for c in &for_node.columns {
            if !self.columns.contains(c) {
                have_all_columns = false;
                break;
            }
        }

        have_all_columns && self.inner.can_reuse_as(&for_node.inner)
    }

    // currently unused
    #[allow(dead_code)]
    pub fn add_ancestor(&mut self, a: MirNodeRef) {
        self.ancestors.push(a)
    }

    pub fn remove_ancestor(&mut self, a: MirNodeRef) {
        match self
            .ancestors
            .iter()
            .position(|x| x.borrow().versioned_name() == a.borrow().versioned_name())
        {
            None => (),
            Some(idx) => {
                self.ancestors.remove(idx);
            }
        }
    }

    pub fn add_child(&mut self, c: MirNodeRef) {
        self.children.push(c)
    }

    pub fn remove_child(&mut self, a: MirNodeRef) {
        match self
            .children
            .iter()
            .position(|x| x.borrow().versioned_name() == a.borrow().versioned_name())
        {
            None => (),
            Some(idx) => {
                self.children.remove(idx);
            }
        }
    }

    pub fn add_column(&mut self, c: Column) {
        match self.inner {
            // the aggregation column must always be the last column
            MirNodeType::Aggregation { .. } => {
                let pos = self.columns.len() - 1;
                self.columns.insert(pos, c.clone());
            }
            _ => self.columns.push(c.clone()),
        }
        self.inner.add_column(c);
    }

    pub fn ancestors(&self) -> &[MirNodeRef] {
        self.ancestors.as_slice()
    }

    pub fn children(&self) -> &[MirNodeRef] {
        self.children.as_slice()
    }

    pub fn columns(&self) -> &[Column] {
        self.columns.as_slice()
    }

    pub fn column_id_for_column(
        &self,
        c: &Column,
        table_mapping: Option<&HashMap<(String, Option<String>), String>>,
    ) -> usize {
        match self.inner {
            // if we're a base, translate to absolute column ID (taking into account deleted
            // columns). We use the column specifications here, which track a tuple of (column
            // spec, absolute column ID).
            // Note that `rposition` is required because multiple columns of the same name might
            // exist if a column has been removed and re-added. We always use the latest column,
            // and assume that only one column of the same name ever exists at the same time.
            MirNodeType::Base {
                ref column_specs, ..
            } => match column_specs
                .iter()
                .rposition(|cs| Column::from(&cs.0.column) == *c)
            {
                None => panic!(
                    "tried to look up non-existent column {:?} in {}",
                    c, self.name
                ),
                Some(id) => column_specs[id]
                    .1
                    .expect("must have an absolute column ID on base"),
            },
            MirNodeType::Reuse { ref node } => node.borrow().column_id_for_column(c, table_mapping),
            // otherwise, just look up in the column set
            _ => match self.columns.iter().position(|cc| cc == c) {
                None => {
                    let get_column_index = |c: &Column, t_name: &str| -> usize {
                        let mut ac = c.clone();
                        ac.table = Some(t_name.to_owned());
                        self.columns
                            .iter()
                            .position(|cc| *cc == ac)
                            .unwrap_or_else(|| {
                                panic!(
                                    "tried to look up non-existent column {:?} on node \
                                     \"{}\" (columns: {:?})",
                                    c, self.name, self.columns
                                )
                            })
                    };
                    // See if table mapping was passed in
                    match table_mapping {
                        // if mapping was passed in, then see if c has an associated table, and check
                        // the mapping for a key based on this
                        Some(map) => match c.table {
                            Some(ref table) => {
                                let key = (c.name.clone(), Some(table.clone()));
                                match map.get(&key) {
                                    Some(ref t_name) => get_column_index(c, t_name),
                                    None => match map.get(&(c.name.clone(), None)) {
                                        Some(ref t_name) => get_column_index(c, t_name),
                                        None => {
                                            panic!("alias is not in mapping table!");
                                        }
                                    },
                                }
                            }
                            None => match map.get(&(c.name.clone(), None)) {
                                Some(ref t_name) => get_column_index(c, t_name),
                                None => panic!(
                                    "tried to look up non-existent column {:?} on node \
                                     \"{}\" (columns: {:?})",
                                    c, self.name, self.columns
                                ),
                            },
                        },
                        // panic if no mapping was passed in
                        None => {
                            panic!(
                                "tried to look up non-existent column {:?} on node \"{}\" \
                                 (columns: {:?})",
                                c, self.name, self.columns
                            );
                        }
                    }
                }
                Some(id) => id,
            },
        }
    }

    pub fn column_specifications(&self) -> &[(ColumnSpecification, Option<usize>)] {
        match self.inner {
            MirNodeType::Base {
                ref column_specs, ..
            } => column_specs.as_slice(),
            _ => panic!("non-base MIR nodes don't have column specifications!"),
        }
    }

    pub fn flow_node_addr(&self) -> Result<NodeIndex, String> {
        match self.flow_node {
            Some(FlowNode::New(na)) | Some(FlowNode::Existing(na)) => Ok(na),
            None => Err(format!(
                "MIR node \"{}\" does not have an associated FlowNode",
                self.versioned_name()
            )),
        }
    }

    #[allow(dead_code)]
    pub fn is_reused(&self) -> bool {
        match self.inner {
            MirNodeType::Reuse { .. } => true,
            _ => false,
        }
    }

    pub fn name(&self) -> &str {
        &self.name
    }

    pub fn referenced_columns(&self) -> Vec<Column> {
        // all projected columns
        let mut columns = self.columns.clone();

        // + any parent columns referenced internally by the operator
        match self.inner {
            MirNodeType::Aggregation { ref on, .. }
            | MirNodeType::Extremum { ref on, .. }
            | MirNodeType::GroupConcat { ref on, .. } => {
                // need the "over" column
                if !columns.contains(on) {
                    columns.push(on.clone());
                }
            }
            MirNodeType::Filter { .. } => {
                let parent = self.ancestors.iter().next().unwrap();
                // need all parent columns
                for c in parent.borrow().columns() {
                    if !columns.contains(&c) {
                        columns.push(c.clone());
                    }
                }
            }
            MirNodeType::Project { ref emit, .. } => {
                for c in emit {
                    if !columns.contains(&c) {
                        columns.push(c.clone());
                    }
                }
            }
            _ => (),
        }
        columns
    }

    pub fn versioned_name(&self) -> String {
        format!("{}_v{}", self.name, self.from_version)
    }

    /// Produce a compact, human-readable description of this node; analogous to the method of the
    /// same name on `Ingredient`.
    fn description(&self) -> String {
        format!(
            "{}: {} / {} columns",
            self.versioned_name(),
            self.inner.description(),
            self.columns.len()
        )
    }
}

/// Specifies the adapatation of an existing base node by column addition/removal.
/// `over` is a `MirNode` of type `Base`.
pub struct BaseNodeAdaptation {
    pub over: MirNodeRef,
    pub columns_added: Vec<ColumnSpecification>,
    pub columns_removed: Vec<ColumnSpecification>,
}

pub enum MirNodeType {
    /// over column, group_by columns
    Aggregation {
        on: Column,
        group_by: Vec<Column>,
        kind: AggregationKind,
    },
    /// column specifications, keys (non-compound), tx flag, adapted base
    Base {
        column_specs: Vec<(ColumnSpecification, Option<usize>)>,
        keys: Vec<Column>,
        adapted_over: Option<BaseNodeAdaptation>,
    },
    /// over column, group_by columns
    Extremum {
        on: Column,
        group_by: Vec<Column>,
        kind: ExtremumKind,
    },
    /// filter conditions (one for each parent column)
    Filter {
        conditions: Vec<Option<FilterCondition>>,
    },
    /// over column, separator
    GroupConcat {
        on: Column,
        separator: String,
    },
    /// no extra info required
    Identity,
    /// left node, right node, on left columns, on right columns, emit columns
    Join {
        on_left: Vec<Column>,
        on_right: Vec<Column>,
        project: Vec<Column>,
    },
    /// on left column, on right column, emit columns
    LeftJoin {
        on_left: Vec<Column>,
        on_right: Vec<Column>,
        project: Vec<Column>,
    },
    /// group columns
    // currently unused
    #[allow(dead_code)]
    Latest {
        group_by: Vec<Column>,
    },
    /// emit columns
    Project {
        emit: Vec<Column>,
        arithmetic: Vec<(String, ArithmeticExpression)>,
        literals: Vec<(String, DataType)>,
    },
    /// emit columns
    Union {
        emit: Vec<Vec<Column>>,
    },
    /// order function, group columns, k
    TopK {
        order: Option<Vec<(Column, OrderType)>>,
        group_by: Vec<Column>,
        k: usize,
        offset: usize,
    },
    // Get the distinct element sorted by a specific column
    Distinct {
        group_by: Vec<Column>,
    },
    /// reuse another node
    Reuse {
        node: MirNodeRef,
    },
    /// leaf (reader) node, keys
    Leaf {
        node: MirNodeRef,
        keys: Vec<Column>,
    },
    /// Rewrite node
    Rewrite {
        value: String,
        column: String,
        key: String,
    },
}

impl MirNodeType {
    fn description(&self) -> String {
        format!("{:?}", self)
    }

    fn add_column(&mut self, c: Column) {
        match *self {
            MirNodeType::Aggregation {
                ref mut group_by, ..
            } => {
                group_by.push(c);
            }
            MirNodeType::Base { .. } => panic!("can't add columns to base nodes!"),
            MirNodeType::Extremum {
                ref mut group_by, ..
            } => {
                group_by.push(c);
            }
            MirNodeType::Filter { ref mut conditions } => {
                conditions.push(None);
            }
            MirNodeType::Join {
                ref mut project, ..
            }
            | MirNodeType::LeftJoin {
                ref mut project, ..
            } => {
                project.push(c);
            }
            MirNodeType::Project { ref mut emit, .. } => {
                emit.push(c);
            }
            MirNodeType::Union { ref mut emit } => {
                for e in emit.iter_mut() {
                    e.push(c.clone());
                }
            }
            MirNodeType::Distinct {
                ref mut group_by, ..
            } => {
                group_by.push(c);
            }
            MirNodeType::TopK {
                ref mut group_by, ..
            } => {
                group_by.push(c);
            }
            _ => (),
        }
    }

    fn can_reuse_as(&self, other: &MirNodeType) -> bool {
        match *self {
            MirNodeType::Reuse { .. } => (), // handled below
            _ => {
                // we're not a `Reuse` ourselves, but the other side might be
                if let MirNodeType::Reuse { ref node } = *other {
                    // it is, so dig deeper
                    // this does not check the projected columns of the inner node for two
                    // reasons:
                    // 1) our own projected columns aren't accessible on `MirNodeType`, but
                    //    only on the outer `MirNode`, which isn't accessible here; but more
                    //    importantly
                    // 2) since this is already a node reuse, the inner, reused node must have
                    //    *at least* a superset of our own (inaccessible) projected columns.
                    // Hence, it is sufficient to check the projected columns on the parent
                    // `MirNode`, and if that check passes, it also holds for the nodes reused
                    // here.
                    return self.can_reuse_as(&node.borrow().inner);
                } else {
                    // handled below
                }
            }
        }

        match *self {
            MirNodeType::Aggregation {
                on: ref our_on,
                group_by: ref our_group_by,
                kind: ref our_kind,
            } => {
                match *other {
                    MirNodeType::Aggregation {
                        ref on,
                        ref group_by,
                        ref kind,
                    } => {
                        // TODO(malte): this is stricter than it needs to be, as it could cover
                        // COUNT-as-SUM-style relationships.
                        our_on == on && our_group_by == group_by && our_kind == kind
                    }
                    _ => false,
                }
            }
            MirNodeType::Base {
                column_specs: ref our_column_specs,
                keys: ref our_keys,
                adapted_over: ref our_adapted_over,
            } => {
                match *other {
                    MirNodeType::Base {
                        ref column_specs,
                        ref keys,
                        ..
                    } => {
                        // if we are instructed to adapt an earlier base node, we cannot reuse
                        // anything directly; we'll have to keep a new MIR node here.
                        if our_adapted_over.is_some() {
                            // TODO(malte): this is a bit more conservative than it needs to be,
                            // since base node adaptation actually *changes* the underlying base
                            // node, so we will actually reuse. However, returning false here
                            // terminates the reuse search unnecessarily. We should handle this
                            // special case.
                            return false;
                        }
                        // note that as long as we are not adapting a previous base node,
                        // we do *not* need `adapted_over` to *match*, since current reuse
                        // does not depend on how base node was created from an earlier one
                        our_column_specs == column_specs && our_keys == keys
                    }
                    _ => false,
                }
            }
            MirNodeType::Extremum {
                on: ref our_on,
                group_by: ref our_group_by,
                kind: ref our_kind,
            } => match *other {
                MirNodeType::Extremum {
                    ref on,
                    ref group_by,
                    ref kind,
                } => our_on == on && our_group_by == group_by && our_kind == kind,
                _ => false,
            },
            MirNodeType::Filter {
                conditions: ref our_conditions,
            } => match *other {
                MirNodeType::Filter { ref conditions } => our_conditions == conditions,
                _ => false,
            },
            MirNodeType::Join {
                on_left: ref our_on_left,
                on_right: ref our_on_right,
                project: ref our_project,
            } => {
                match *other {
                    MirNodeType::Join {
                        ref on_left,
                        ref on_right,
                        ref project,
                    } => {
                        // TODO(malte): column order does not actually need to match, but this only
                        // succeeds if it does.
                        our_on_left == on_left && our_on_right == on_right && our_project == project
                    }
                    _ => false,
                }
            }
            MirNodeType::LeftJoin {
                on_left: ref our_on_left,
                on_right: ref our_on_right,
                project: ref our_project,
            } => {
                match *other {
                    MirNodeType::LeftJoin {
                        ref on_left,
                        ref on_right,
                        ref project,
                    } => {
                        // TODO(malte): column order does not actually need to match, but this only
                        // succeeds if it does.
                        our_on_left == on_left && our_on_right == on_right && our_project == project
                    }
                    _ => false,
                }
            }
            MirNodeType::Project {
                emit: ref our_emit,
                literals: ref our_literals,
                arithmetic: ref our_arithmetic,
            } => match *other {
                MirNodeType::Project {
                    ref emit,
                    ref literals,
                    ref arithmetic,
                } => our_emit == emit && our_literals == literals && our_arithmetic == arithmetic,
                _ => false,
            },
            MirNodeType::Distinct {
                group_by: ref our_group_by,
            } => match *other {
                MirNodeType::Distinct { ref group_by } => group_by == our_group_by,
                _ => false,
            },
            MirNodeType::Reuse { node: ref us } => {
                match *other {
                    // both nodes are `Reuse` nodes, so we simply compare the both sides' reuse
                    // target
                    MirNodeType::Reuse { ref node } => us.borrow().can_reuse_as(&*node.borrow()),
                    // we're a `Reuse`, the other side isn't, so see if our reuse target's `inner`
                    // can be reused for the other side. It's sufficient to check the target's
                    // `inner` because reuse implies that our target has at least a superset of our
                    // projected columns (see earlier comment).
                    _ => us.borrow().inner.can_reuse_as(other),
                }
            }
            MirNodeType::TopK {
                order: ref our_order,
                group_by: ref our_group_by,
                k: our_k,
                offset: our_offset,
            } => match *other {
                MirNodeType::TopK {
                    ref order,
                    ref group_by,
                    k,
                    offset,
                } => {
                    order == our_order
                        && group_by == our_group_by
                        && k == our_k
                        && offset == our_offset
                }
                _ => false,
            },
            MirNodeType::Leaf {
                keys: ref our_keys, ..
            } => match *other {
                MirNodeType::Leaf { ref keys, .. } => keys == our_keys,
                _ => false,
            },
            MirNodeType::Union { emit: ref our_emit } => match *other {
                MirNodeType::Union { ref emit } => emit == our_emit,
                _ => false,
            },
            MirNodeType::Rewrite {
                value: ref our_value,
                key: ref our_key,
                column: ref our_col,
            } => match *other {
                MirNodeType::Rewrite {
                    ref value,
                    ref key,
                    ref column,
                } => (value == our_value && our_key == key && our_col == column),
                _ => false,
            },
            _ => unimplemented!(),
        }
    }
}

impl Display for MirNode {
    fn fmt(&self, f: &mut Formatter) -> Result<(), Error> {
        write!(f, "{}", self.inner.description())
    }
}

impl Debug for MirNode {
    fn fmt(&self, f: &mut Formatter) -> Result<(), Error> {
        write!(
            f,
            "{}, {} ancestors ({}), {} children ({})",
            self.description(),
            self.ancestors.len(),
            self.ancestors
                .iter()
                .map(|a| a.borrow().versioned_name())
                .collect::<Vec<_>>()
                .join(", "),
            self.children.len(),
            self.children
                .iter()
                .map(|c| c.borrow().versioned_name())
                .collect::<Vec<_>>()
                .join(", ")
        )
    }
}

impl Debug for MirNodeType {
    fn fmt(&self, f: &mut Formatter) -> Result<(), Error> {
        match *self {
            MirNodeType::Aggregation {
                ref on,
                ref group_by,
                ref kind,
            } => {
                let op_string = match *kind {
                    AggregationKind::COUNT => format!("|*|({})", on.name.as_str()),
                    AggregationKind::SUM => format!("𝛴({})", on.name.as_str()),
                };
                let group_cols = group_by
                    .iter()
                    .map(|c| c.name.as_str())
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(f, "{} γ[{}]", op_string, group_cols)
            }
            MirNodeType::Base {
                ref column_specs,
                ref keys,
                ..
            } => write!(
                f,
                "B [{}; ⚷: {}]",
                column_specs
                    .iter()
                    .map(|&(ref cs, _)| cs.column.name.as_str())
                    .collect::<Vec<_>>()
                    .join(", "),
                keys.iter()
                    .map(|c| c.name.as_str())
                    .collect::<Vec<_>>()
                    .join(", ")
            ),
            MirNodeType::Extremum {
                ref on,
                ref group_by,
                ref kind,
            } => {
                let op_string = match *kind {
                    ExtremumKind::MIN => format!("min({})", on.name.as_str()),
                    ExtremumKind::MAX => format!("max({})", on.name.as_str()),
                };
                let group_cols = group_by
                    .iter()
                    .map(|c| c.name.as_str())
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(f, "{} γ[{}]", op_string, group_cols)
            }
            MirNodeType::Filter { ref conditions } => {
                use regex::Regex;

                let escape = |s: &str| {
                    Regex::new("([<>])")
                        .unwrap()
                        .replace_all(s, "\\$1")
                        .to_string()
                };
                write!(
                    f,
                    "σ[{}]",
                    conditions
                        .iter()
                        .enumerate()
                        .filter_map(|(i, ref e)| match e.as_ref() {
                            Some(cond) => match *cond {
                                FilterCondition::Comparison(ref op, ref x) => {
                                    Some(format!("f{} {} {:?}", i, escape(&format!("{}", op)), x))
                                }
                                FilterCondition::In(ref xs) => Some(format!(
                                    "f{} IN ({})",
                                    i,
                                    xs.iter()
                                        .map(|d| format!("{}", d))
                                        .collect::<Vec<_>>()
                                        .join(", ")
                                )),
                            },
                            None => None,
                        })
                        .collect::<Vec<_>>()
                        .as_slice()
                        .join(", ")
                )
            }
            MirNodeType::GroupConcat {
                ref on,
                ref separator,
            } => write!(f, "||([{}], \"{}\")", on.name, separator),
            MirNodeType::Identity => write!(f, "≡"),
            MirNodeType::Join {
                ref on_left,
                ref on_right,
                ref project,
            } => {
                let jc = on_left
                    .iter()
                    .zip(on_right)
                    .map(|(l, r)| format!("{}:{}", l.name, r.name))
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(
                    f,
                    "⋈ [{} on {}]",
                    project
                        .iter()
                        .map(|c| c.name.as_str())
                        .collect::<Vec<_>>()
                        .join(", "),
                    jc
                )
            }
            MirNodeType::Leaf { ref keys, .. } => {
                let key_cols = keys
                    .iter()
                    .map(|k| k.name.clone())
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(f, "Leaf [⚷: {}]", key_cols)
            }
            MirNodeType::LeftJoin {
                ref on_left,
                ref on_right,
                ref project,
            } => {
                let jc = on_left
                    .iter()
                    .zip(on_right)
                    .map(|(l, r)| format!("{}:{}", l.name, r.name))
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(
                    f,
                    "⋉ [{} on {}]",
                    project
                        .iter()
                        .map(|c| c.name.as_str())
                        .collect::<Vec<_>>()
                        .join(", "),
                    jc
                )
            }
            MirNodeType::Latest { ref group_by } => {
                let key_cols = group_by
                    .iter()
                    .map(|k| k.name.clone())
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(f, "⧖ γ[{}]", key_cols)
            }
            MirNodeType::Project {
                ref emit,
                ref literals,
                ref arithmetic,
            } => write!(
                f,
                "π [{}{}{}]",
                emit.iter()
                    .map(|c| c.name.as_str())
                    .collect::<Vec<_>>()
                    .join(", "),
                if arithmetic.is_empty() {
                    "".into()
                } else {
                    format!(
                        ", {}",
                        arithmetic
                            .iter()
                            .map(|&(ref n, ref e)| format!("{}: {}", n, e))
                            .collect::<Vec<_>>()
                            .join(", ")
                    )
                },
                if literals.is_empty() {
                    "".into()
                } else {
                    format!(
                        ", lit: {}",
                        literals
                            .iter()
                            .map(|&(ref n, ref v)| format!("{}: {}", n, v))
                            .collect::<Vec<_>>()
                            .join(", ")
                    )
                },
            ),
            MirNodeType::Reuse { ref node } => write!(
                f,
                "Reuse [{}: {}]",
                node.borrow().versioned_name(),
                node.borrow()
            ),
            MirNodeType::Distinct { ref group_by } => {
                let key_cols = group_by
                    .iter()
                    .map(|k| k.name.clone())
                    .collect::<Vec<_>>()
                    .join(", ");
                write!(f, "Distinct [γ: {}]", key_cols)
            }
            MirNodeType::TopK {
                ref order, ref k, ..
            } => write!(f, "TopK [k: {}, {:?}]", k, order),
            MirNodeType::Union { ref emit } => {
                let cols = emit
                    .iter()
                    .map(|c| {
                        c.iter()
                            .map(|e| e.name.clone())
                            .collect::<Vec<_>>()
                            .join(", ")
                    })
                    .collect::<Vec<_>>()
                    .join(" ⋃ ");

                write!(f, "{}", cols)
            }
            MirNodeType::Rewrite { ref column, .. } => write!(f, "Rw [{}]", column),
        }
    }
}
